import ConfigParser
import base64
import os
import urlparse
import subprocess
import yaml

from charmhelpers import fetch
from charmhelpers.core import (
    hookenv,
    host,
)
from charmhelpers.core.services import helpers
from charmhelpers.payload import (archive, execd)

from charmhelpers.contrib.charmsupport import nrpe

REQUIRED_PACKAGES = [
    'python-virtualenv', 'python3-dev', 'libpq5'
]

WSGI_USER = 'www-data'
WSGI_GROUP = 'www-data'

config = hookenv.config()

BASE_DIR = '/srv/spec-manager/{}'.format(config['environment'])
SERVICE_DIR = os.path.join(BASE_DIR, 'spec-manager')
LOG_DIR = os.path.join(BASE_DIR, 'logs')
ETC_DIR = os.path.join(BASE_DIR, 'etc')


def log_start(service_name):
    hookenv.log('spec-manager starting')


def ensure_directories(service_name):
    for dirpath in [BASE_DIR, ETC_DIR]:
        host.mkdir(dirpath, owner='root', group='ubuntu')


def basenode(service_name):
    hookenv.log("Executing basenode")
    execd.execd_preinstall()


def install_packages(service_name):
    hookenv.log('Installing dependencies...')
    fetch.configure_sources(update=True)
    fetch.apt_install(REQUIRED_PACKAGES, fatal=True)


def get_cloud_service_from_tarball(service_name):
    files_dir = os.path.join(hookenv.charm_dir(), 'files')
    tarball = os.path.join(files_dir, 'spec-manager.tgz')
    if not os.path.exists(SERVICE_DIR):
        hookenv.log('Installing the code for the first time from tarball')
        archive.extract_tarfile(tarball, os.path.dirname(SERVICE_DIR))


def get_config_file(service_name):
    config_file = config.get('config-file')
    if config_file is not None:
        with open(os.path.join(SERVICE_DIR, 'service.conf'), 'w') as f:
            f.write(base64.b64decode(config_file))


def install_python_packages(service_name):
    env_dir = os.path.join(SERVICE_DIR, 'env')
    if os.path.exists(env_dir):
        hookenv.log('Service venv already exists, nothing to do ...')
        return

    hookenv.log('Installing python packages into the venv ...')
    subprocess.check_call(['virtualenv', '-p', 'python3', env_dir])

    pip_cache = os.path.join(hookenv.charm_dir(), 'files', 'pip-cache')
    requirements = os.path.join(SERVICE_DIR, 'requirements.txt')

    hookenv.log('Installing from download cache.')
    subprocess.check_call(['%s/bin/pip' % env_dir,
                           'install',
                           '--no-index',
                           '--find-links={}'.format(pip_cache),
                           '-r', requirements])


def prepare_for_wsgi(service_name):
    host.adduser(WSGI_USER)
    host.add_group(WSGI_GROUP)
    host.add_user_to_group(WSGI_USER, WSGI_GROUP)
    if not os.path.exists(LOG_DIR):
        host.mkdir(LOG_DIR, WSGI_USER, WSGI_GROUP, perms=0o755)


def publish_conn_check_relation_data(service_name):
    cr = ConnCheckRelation()
    for rel in hookenv.relations_of_type('conn-check'):
        hookenv.relation_set(rel['__relid__'], **cr.provide_data())


class WebsiteRelation(helpers.HttpRelation):

    def provide_data(self):
        hookenv.log('Providing Website relation data.')
        data = super(WebsiteRelation, self).provide_data()
        data['port'] = 8000
        return data


def write_db_config(service_name):
    db_rels = hookenv.relations_of_type('db')
    db_config_created = False
    db_config_path = os.path.join(ETC_DIR, 'db.yaml')
    db_schema_config_path = os.path.join(ETC_DIR, 'db-schema.yaml')
    db_admin_config_created = False
    db_admin_config_path = os.path.join(ETC_DIR, 'db-admin.yaml')
    if db_rels:
        # use the master/standalone server
        for db_rel in db_rels:
            if 'user' not in db_rel or 'state' not in db_rel:
                continue
            if db_rel['state'] in ('standalone', 'master'):
                app_db_config = {}
                for k in ('database', 'user', 'password', 'host', 'port'):
                    app_db_config[k] = db_rel[k]
                host.write_file(db_config_path, yaml.dump(app_db_config),
                                owner='root', group='www-data', perms=0o440)
                schema_db_config = {}
                for k in ('database', 'schema_user', 'schema_password',
                          'host', 'port'):
                    schema_db_config[k] = db_rel[k]
                host.write_file(db_schema_config_path, yaml.dump(schema_db_config),
                                owner='root', group='www-data', perms=0o440)
                db_config_created = True
                # update service.conf inplace
                parser = ConfigParser.ConfigParser()
                config_file = os.path.join(SERVICE_DIR, 'service.conf')
                parser.read([config_file])
                parser.set('db', 'host', value=db_rel['host'])
                parser.set('db', 'username', value=db_rel['user'])
                parser.set('db', 'password', value=db_rel['password'])
                parser.set('db', 'port', value=db_rel['port'])
                with open(config_file, 'w') as fp:
                    parser.write(fp)
                    fp.flush()
    # cleanup if nothing was created
    if not db_config_created:
        if os.path.exists(db_config_path):
            os.remove(db_config_path)
        if os.path.exists(db_schema_config_path):
            os.remove(db_schema_config_path)
    db_admin_rels = hookenv.relations_of_type('db-admin')
    if db_admin_rels and db_config_created:
        # use the master/standalone server
        for db_rel in db_admin_rels:
            if 'user' not in db_rel or 'state' not in db_rel:
                continue
            if db_rel['state'] in ('standalone', 'master'):
                content = yaml.dump(db_rel)
                host.write_file(db_admin_config_path, content,
                                owner='root', group='root', perms=0o400)
                db_admin_config_created = True

    # cleanup if nothing was updated
    if not db_config_created and os.path.exists(db_config_path):
        os.remove(db_config_path)
    if not db_admin_config_created and os.path.exists(db_admin_config_path):
        os.remove(db_admin_config_path)


class PostgresqlRelation(helpers.RelationContext):
    """
    Relation context for the `postgresql` interface.

    :param str name: Override the relation :attr:`name`, since it can vary from charm to charm
    :param list additional_required_keys: Extend the list of :attr:`required_keys`
    """
    name = 'db'
    interface = 'pgsql'
    required_keys = []
    roles = None
    database = None
    user = None

    def __init__(self, required=False):
        config = hookenv.config()
        self.roles = config['db_roles']
        self.database = config['db_name']
        self.user = config['db_user']
        if required:
            self.required_keys = ['host', 'user', 'password', 'database']
        super(PostgresqlRelation, self).__init__(name=self.name)

    def is_ready(self):
        data = self.get(self.name, [])
        if not data:
            return False
        all_rels = []
        for d in data:
            if d.get('database') == self.database and self._unit_in(d.get('allowed-units', [])):
                all_rels.append(True)
            else:
                all_rels.append(False)
        return all(all_rels)

    def provide_data(self):
        return dict(roles=self.roles, database=self.database, user=self.user)

    def _unit_in(self, units):
        return hookenv.local_unit() in units


class ConnCheckRelation(helpers.RelationContext):

    name = 'conn-check'
    interface = 'conn-check'
    required_keys = []

    def provide_data(self):
        config = hookenv.config()
        conn_check_config = []
        base64_config = config.get('conn_check_config')
        if base64_config:
            conn_check_config.extend(yaml.safe_load(base64.b64decode(base64_config)))
        conn_check_config.extend(self._get_postgresql_checks())
        conn_check_config.extend(self._get_amqp_checks())
        # save the yaml conn-check config file
        content = yaml.dump(conn_check_config, default_flow_style=False)
        conn_check_config_path = os.path.join(ETC_DIR, 'conn-check-config.yaml')
        host.write_file(conn_check_config_path, content,
                        owner='ubuntu', group='nagios', perms=0o440)
        # notify the relation
        return dict(config_path=conn_check_config_path,
                    nagios_servicegroups=config['nagios_context'])

    def _get_postgresql_checks(self):
        config = hookenv.config()
        checks = []
        db_rels = hookenv.relations_of_type('db')
        required_data = ('port', 'host', 'user', 'password', 'database')
        for server in db_rels:
            if len(required_data) != len([k for k in required_data if k in server]) \
                    and server['database'] == config['db_name']:
                continue
            db_check = {'type': 'postgres',
                        'host': server['host'],
                        'port': int(server['port']),
                        'username': server['user'],
                        'password': server['password'],
                        'database': server['database']}
            checks.append(db_check)
        return checks

    def _get_amqp_checks(self):
        checks = []
        # read amqp uris from config file to build amqp check
        parser = ConfigParser.ConfigParser()
        config_file = os.path.join(SERVICE_DIR, 'service.conf')
        parser.read([config_file])
        if 'amqp' not in parser.sections():
            return checks
        # only if amqp section is defined
        for amqp_uri in parser.get('amqp', 'uris').split():
            parsed = urlparse.urlparse(amqp_uri)
            vhost = urlparse.unquote(parsed.path[1:]) or '/'
            use_tls = parsed.scheme == "amqps"
            amqp_check = dict(type="amqp",
                              host=parsed.hostname,
                              port=parsed.port,
                              username=parsed.username,
                              password=parsed.password,
                              use_tls=use_tls,
                            vhost=vhost)
            checks.append(amqp_check)
        return checks


def nrpe_external_master_relation(service_name):
    ''' Configure the nrpe-external-master relation '''
    nrpe_compat = nrpe.NRPE()

    nrpe_compat.add_check(
        shortname="check_app_http_rabbit",
        description="Application and app rabbitmq connection.",
        check_cmd="/usr/lib/nagios/plugins/check_http -I 127.0.0.1 -p 8000 -e ' 200 OK' -u '/nagios/rabbitmq'",
    )
    nrpe_compat.add_check(
        shortname="check_app_http_postgres",
        description="Application and app postgres connection.",
        check_cmd="/usr/lib/nagios/plugins/check_http -I 127.0.0.1 -p 8000 -e ' 200 OK' -u '/nagios/postgres'",
    )

    nrpe_compat.write()
